@@ -15,7 +15,6 @@ from libpipe import futils
 from libpipe.attrmap import OrderedSet
 
 # Lazy loaded:
-# - libpipe.msutils in ObsIDMeta.set_ms_info()
 # - .obsdata.ObsData in VisRevision.get_data()
 
 from .settings import Settings
@@ -28,7 +27,7 @@ OBS_VIS_REV_JSON = 'vis_list.json'
 IMG_REV_TOML = 'img_config.toml'
 OBS_VIS_REV_TOML = 'vis_config.toml'
 
-all_stokes = ['I', 'V', 'Q', 'U']
+all_stokes = ['I', 'V', 'Q', 'U', 'XX', 'YY', 'XY', 'YX']
 
 obs_id_info_keys = ['start_time', 'end_time', 'int_time', 'total_time',
                     'n_channels', 'chan_width', 'is_sph', 'is_combined']
@@ -199,6 +198,10 @@ class ImageRevision(object):
     def list_obs_ids(self):
         return listdir_isdir(os.path.join(self.get_path(), 'images'))
 
+    def filter_obs_ids(self, obs_id_str):
+        all_obs_ids = self.list_obs_ids()
+        return set(obs_id for k in obs_id_str.split(',') for obs_id in fnmatch.filter(all_obs_ids, k))
+
     def has_obs_id(self, obs_id):
         return obs_id in self.list_obs_ids()
 
@@ -286,6 +289,10 @@ class VisRevision(object):
     def list_obs_ids(self):
         return listdir_isdir(self.get_path(), ['results', 'combined_obs_ids'])
 
+    def filter_obs_ids(self, obs_id_str):
+        all_obs_ids = self.list_obs_ids()
+        return set(obs_id for k in obs_id_str.split(',') for obs_id in fnmatch.filter(all_obs_ids, k))
+
     def has_obs_id(self, obs_id):
         return obs_id in self.list_obs_ids()
 
@@ -296,6 +303,9 @@ class VisRevision(object):
 
         return img_meta
 
+    def get_lst_bins(self, obs_id, return_time_index=False):
+        return ImageRevision(self.settings).get_lst_bins(obs_id, return_time_index=return_time_index)
+
     def get_meta(self, obs_id, lst_idx=None):
         meta_file = self.get_meta_path(obs_id, lst_idx=lst_idx)
         if os.path.exists(meta_file):
@@ -305,6 +315,8 @@ class VisRevision(object):
     def get_data(self, obs_id, lst_idx=None):
         from .obsdata import ObsData
 
+        assert self.has_obs_id(obs_id)
+
         return ObsData(self, obs_id, lst_idx=lst_idx)
 
     def get_h5_file(self, obs_id, stokes, even=False, odd=False, lst_idx=None):
@@ -352,9 +364,7 @@ class ObsIDMeta(AbstractJsonMeta):
         return ObsIDMeta(file, d)
 
     def set_ms_files(self, ms_files):
-        from libpipe import msutils
-
-        self.update(msutils.get_info_from_ms_files(ms_files[0]))
+        self.update(utils.get_info_from_ms_files(ms_files[0]))
         self['ms_files'] = ms_files
         self.save()
 
