@@ -17,7 +17,7 @@ from . import utils
 
 TEMPLATE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
 
-TEMPLATE_NAMES = ['a12', 'nenufar', 'hba']
+TEMPLATE_NAMES = ['a12', 'nenufar', 'hba', 'alo']
 
 
 class Settings(BaseSettings):
@@ -31,10 +31,10 @@ class Settings(BaseSettings):
         val_fct(u.Unit(self.image.scale).is_equivalent((u.rad, u.s)), 'scale')
         val_fct(len(self.image.size.split()) == 2 and all(k.isdigit() for k in self.image.size.split()), 'size')
         val_fct(self.image.channels_out == 'all' or str(self.image.channels_out).startswith(
-            'every') or self.image.channels_out.isdigit(), 'channels_out')
+            'every') or utils.is_number(self.image.channels_out), 'channels_out')
         val_fct(isinstance(self.image.umin, (int, float)), 'umin')
         val_fct(isinstance(self.image.umax, (int, float)), 'umax')
-        val_fct(utils.all_in_other(set(self.image.stokes), set('IQUV')), 'stokes')
+        val_fct(utils.all_in_other(set(self.image.stokes), set('IQUVXY')), 'stokes')
         val_fct(isinstance(self.image.split_even_odd, bool), 'split_even_odd')
         val_fct(isinstance(self.image.time_start_index, int), 'time_start_index')
         val_fct(isinstance(self.image.time_end_index, int), 'time_end_index')
@@ -45,6 +45,7 @@ class Settings(BaseSettings):
         val_fct(isinstance(self.image.clean_auto_threshold, (int, float)), 'clean_auto_threshold')
         val_fct(isinstance(self.image.clean_mgain, (int, float)), 'clean_mgain')
         val_fct(isinstance(self.image.clean_residual, bool), 'clean_residual')
+        val_fct(isinstance(self.image.clean_spectra_fit_nterms, int), 'clean_spectra_fit_nterms')
 
     def validate_worker(self, val_fct):
         val_fct(isinstance(self.worker.max_concurrent, int), 'max_concurrent')
@@ -86,6 +87,7 @@ class Settings(BaseSettings):
     def validate_gpr(self, val_fct):
         val_fct(futils.is_file(self.gpr.config_i), 'config_i')
         val_fct(futils.is_file(self.gpr.config_v), 'config_v')
+        val_fct(isinstance(self.gpr.plot_results, bool), 'plot_results')
 
     def validate_ssins(self, val_fct):
         val_fct(isinstance(self.ssins.n_time_avg, list), 'n_time_avg')
@@ -95,12 +97,12 @@ class Settings(BaseSettings):
 
     def validate_delay_flagger(self, val_fct):
         for k in ['umin', 'umax', 'n_sigma_i', 'n_sigma_v']:
-            val_fct(isinstance(self.ssins[k], (int, float)), k)
+            val_fct(isinstance(self.delay_flagger[k], (int, float)), k)
         for k in ['n_time_avg', 'n_times']:
-            val_fct(isinstance(self.ssins[k], int), k)
+            val_fct(isinstance(self.delay_flagger[k], int), k)
 
 
 def get_template_path(template_name):
-    assert template_name in TEMPLATE_NAMES
+    assert template_name in TEMPLATE_NAMES, f'template_name should be one of: {" ".join(TEMPLATE_NAMES)}'   
 
     return os.path.join(TEMPLATE_DIR, f'default_{template_name}.toml')
