@@ -5,6 +5,9 @@
 import os
 import re
 import glob
+import tempfile
+import datetime
+import itertools
 
 import numpy as np
 import astropy.io.fits as pf
@@ -33,6 +36,15 @@ def get_all_tasks_descriptions():
     return d
 
 
+def get_target_host(in_file, worker_settings):
+    host = None
+    if worker_settings.run_on_file_host and worker_settings.run_on_file_host_pattern:
+        r = re.search(worker_settings.run_on_file_host_pattern, in_file)
+        if r is not None:
+            host = r.group(1)
+    return host
+
+
 class AbstractSubTask(object):
 
     def __init__(self, settings_obj, exec_name):
@@ -61,17 +73,24 @@ class AbstractTask(object):
         self.name = name
         self.settings = settings_obj.copy()
         self.settings.worker.update(worker_kargs)
+        s_worker = self.settings.worker
 
         if not self.settings.validate('worker'):
             raise ValueError('Invalid worker settings.')
 
-        self.worker = worker.get_worker_pool(name, **self.settings.worker)
+        self.worker = worker.get_worker_pool(name, nodes=s_worker.nodes, max_concurrent=s_worker.max_concurrent,
+                                             env_file=s_worker.env_file)
 
     def init_for_new_task(self, post_task_name):
-        self.worker = worker.get_worker_pool(post_task_name, **self.settings.worker)
+        s_worker = self.settings.worker
+        self.worker = worker.get_worker_pool(post_task_name, nodes=s_worker.nodes,
+                                             max_concurrent=s_worker.max_concurrent,
+                                             env_file=s_worker.env_file)
 
-    def add_sub_task(self, sub_task):
+    def add_sub_task(self, sub_task, run_on_host=None):
         log_file = sub_task.get_log_file()
+        if log_file:
+            log_file = log_file.rstrip('/')
 
         if log_file is not None:
             if not os.path.exists(os.path.dirname(log_file)):
@@ -80,7 +99,7 @@ class AbstractTask(object):
         cmd = sub_task.build_command()
         if self.settings.debug:
             print(cmd)
-        self.worker.add(cmd, output_file=log_file, done_callback=sub_task.done_callback)
+        self.worker.add(cmd, output_file=log_file, done_callback=sub_task.done_callback, run_on_host=run_on_host)
 
     def run(self, obs_ids):
         raise NotImplementedError()
@@ -99,14 +118,17 @@ class AbstractTask(object):
 
 class ImagingSubTask(AbstractSubTask):
 
-    def __init__(self, settings_obj, obs_id, inp_ms_files, even=False, odd=False):
+    def __init__(self, settings_obj, obs_id, inp_ms_files, even=False, odd=False, index=0):
         AbstractSubTask.__init__(self, settings_obj, 'wsclean')
         self.inp_ms_files = inp_ms_files
         self.obs_id = obs_id
         self.even = even
         self.odd = odd
+        self.index = index
         self.img_rev = database.ImageRevision(self.settings)
         self.obs_meta = database.Database(self.settings).get_obs_id_meta(self.obs_id)
+        self.temp_dir = None
+        self.temp_protect_file = None
 
     def get_dir(self):
         return self.img_rev.get_image_path(self.obs_id)
@@ -115,10 +137,12 @@ class ImagingSubTask(AbstractSubTask):
         res = re.search(r'_SB(\d+)', os.path.basename(self.inp_ms_files[0]))
         if res is not None:
             return res.groups()[0]
-        return '000'
+        else:
+            return f'{self.index:04d}'
 
     def get_log_file(self):
-        return os.path.join(self.get_dir(), 'logs', os.path.basename(self.get_fits_prefix()) + '.log')
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.get_dir(), 'logs', os.path.basename(self.get_fits_prefix()) + f'_{date_str}.log')
 
     def get_fits_prefix(self):
         img = self.settings.image
@@ -165,13 +189,22 @@ class ImagingSubTask(AbstractSubTask):
                 files = glob.glob(f'{self.get_fits_prefix()}*-t{i:04}-*.fits')
                 self.register_files(files, lst_idx=data_lst_idx, img_name=img_name)
 
+        try:
+            os.remove(self.temp_protect_file)
+        except (FileNotFoundError, OSError):
+            # ignore
+            pass
+
+        try:
+            os.rmdir(self.temp_dir)
+        except (FileNotFoundError, OSError):
+            # ignore
+            pass
+
     def get_parameters(self):
         args = []
         img = self.settings.image
         # Default arguments
-        temp_dir = os.path.join(self.get_dir(), 'tmp' + database.even_odd_str(self.even, self.odd))
-        if not os.path.isdir(temp_dir):
-            os.makedirs(temp_dir)
 
         channels_out = img.channels_out
         if channels_out == 'all':
@@ -184,9 +217,23 @@ class ImagingSubTask(AbstractSubTask):
             channels_out = self.obs_meta.n_channels // int(channels_out[5:])
         channels_out = int(channels_out)
 
-        args.append("-oversampling 4095 -kernel-size 15 -nwlayers 32 -grid-mode kb -taper-edge 100")
-        args.append("-padding 2 -visibility-weighting-mode unit -reorder")
-        args.append(f'-temp-dir {temp_dir} -make-psf -no-dirty')
+        args.append(f"-oversampling {img.oversampling} -kernel-size {img.kernel_size} -nwlayers {img.nwlayers}")
+        args.append(f"-grid-mode kb -taper-edge {img.taper_edge} -padding {img.padding}")
+        args.append(f"-visibility-weighting-mode {img.visibility_weighting_mode} -reorder -make-psf -no-dirty")
+
+        # If MS directory is not writable, we need to define a temp directory
+        if not all(os.access(ms_file, os.W_OK) for ms_file in self.inp_ms_files):
+            self.temp_dir = os.path.join(self.get_dir(), 'tmp' + database.even_odd_str(self.even, self.odd))
+        else:
+            self.temp_dir = os.path.join(self.inp_ms_files[0], 'tmp' + database.even_odd_str(self.even, self.odd))
+
+        if not os.path.isdir(self.temp_dir):
+            os.makedirs(self.temp_dir)
+
+        # Make sure we do not remove this directory by accident
+        _, self.temp_protect_file = tempfile.mkstemp(dir=self.temp_dir)
+
+        args.append(f'-temp-dir {self.temp_dir}')
 
         if self.even:
             args.append('-even-timesteps')
@@ -209,6 +256,7 @@ class ImagingSubTask(AbstractSubTask):
         if img.clean_niter > 0:
             args.append(f'-niter {img.clean_niter} -auto-threshold {img.clean_auto_threshold}')
             args.append(f'-local-rms -mgain {img.clean_mgain} -stop-negative -no-update-model-required')
+            args.append(f'-join-channels -fit-spectral-pol {img.clean_spectra_fit_nterms}')
 
             if channels_out > 1:
                 args.append('-join-channels')
@@ -258,12 +306,16 @@ class ImagingTask(AbstractTask):
                 for i in self.img_rev.get_lst_bins(obs_id).keys():
                     self.img_rev.new_obs_id(obs_id, lst_idx=i)
 
-            for ms_files_channel in ms_files:
+            for i, ms_files_channel in enumerate(ms_files):
+                target_host = get_target_host(ms_files_channel[0], self.settings.worker)
                 if self.settings.image.split_even_odd:
-                    self.add_sub_task(ImagingSubTask(self.settings, obs_id, ms_files_channel, even=True))
-                    self.add_sub_task(ImagingSubTask(self.settings, obs_id, ms_files_channel, odd=True))
+                    self.add_sub_task(ImagingSubTask(self.settings, obs_id, ms_files_channel, even=True, index=i),
+                                      run_on_host=target_host)
+                    self.add_sub_task(ImagingSubTask(self.settings, obs_id, ms_files_channel, odd=True, index=i),
+                                      run_on_host=target_host)
                 else:
-                    self.add_sub_task(ImagingSubTask(self.settings, obs_id, ms_files_channel))
+                    self.add_sub_task(ImagingSubTask(self.settings, obs_id, ms_files_channel, index=i),
+                                      run_on_host=target_host)
 
             all_obs_id_out.append(obs_id)
 
@@ -300,8 +352,9 @@ class GenVisCubeSubTask(AbstractSubTask):
         return self.vis_rev.get_h5_file(self.obs_id, self.stokes.upper(), self.even, self.odd, lst_idx=self.lst_idx)
 
     def get_log_file(self):
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
         return os.path.join(self.vis_rev.get_h5_path(self.obs_id, self.lst_idx), 'logs',
-                            os.path.basename(self.get_out_file().replace('.h5', '.log')))
+                            os.path.basename(self.get_out_file().replace('.h5', f'_{date_str}.log')))
 
     def done_callback(self):
         meta = self.vis_rev.get_meta(self.obs_id, self.lst_idx)
@@ -410,10 +463,86 @@ class GenVisCubeTask(AbstractTask):
         return all_obs_id_out
 
 
+class ConcatMSSubTask(AbstractSubTask):
+    """
+    Subtask that calls the CLI 'mstool concat <MS...> <OUT.MS>'
+    to concatenate MS files in time.
+    """
+
+    def __init__(self, settings_obj, msfiles_flat, obs_id_out):
+        super().__init__(settings_obj, 'mstool')
+        self.db = database.Database(self.settings)
+        # preserve given order (assumed time order already)
+        self.msfiles_flat = list(msfiles_flat)
+        self.obs_id_out = obs_id_out
+
+    def get_dir(self):
+        return os.path.join(self.settings.get_path('data_dir'),
+                            'concat_ms', self.obs_id_out)
+
+    def get_out_file(self):
+        return os.path.join(self.get_dir(), f'{self.obs_id_out}.MS')
+
+    def get_log_file(self):
+        ts = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return self.get_out_file().replace('.MS', f'_{ts}.log')
+
+    def done_callback(self):
+        meta = self.db.new_obs_id(self.obs_id_out) if not self.db.has_obs_id(self.obs_id_out) \
+              else self.db.get_obs_id_meta(self.obs_id_out)
+        meta.set_ms_files([[self.get_out_file()]])
+
+    def get_parameters(self):
+        # Ensure parent dir exists; mstool will create the MS dir itself
+        os.makedirs(self.get_dir(), exist_ok=True)
+
+        # Command: mstool concat <MS1.MS> <MS2.MS> ... <OUT.MS>
+        cmd = ['concat', *self.msfiles_flat, self.get_out_file()]
+        return cmd
+
+
+class ConcatMSTask(AbstractTask):
+
+    name = 'concat_ms'
+    desc = 'Concatenate MS files in time, create a new obs_id'
+
+    def __init__(self, settings_obj):
+        super().__init__('ConcatMS', settings_obj, max_concurrent=1)
+        self.db = database.Database(self.settings)
+
+    def run(self, obs_ids):
+        all_obs_id_out = []
+
+        for obs_id in obs_ids:
+            if not self.db.has_obs_id(obs_id):
+                print(f'Error: Obs ID {obs_id} does not exist.')
+                return []
+
+            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
+
+            # Must be exactly one inner list with all MS
+            if not ms_files or len(ms_files) != 1:
+                print(f'Error: Obs ID {obs_id} must have exactly one ms_files entry with all MS to concat.')
+                return []
+
+            msfiles_flat = ms_files[0]
+
+            if len(msfiles_flat) < 2:
+                print(f'Error: Obs ID {obs_id} needs at least 2 MS to concatenate (found {len(msfiles_flat)}).')
+                return []
+
+            obs_id_out = f'{obs_id}_concat'.replace(' ', '')
+            self.add_sub_task(ConcatMSSubTask(self.settings, msfiles_flat, obs_id_out))
+            all_obs_id_out.append(obs_id_out)
+
+        self.execute()
+        return all_obs_id_out
+
+
 class MergeMSSubTask(AbstractSubTask):
 
     def __init__(self, settings_obj, ms_files, obs_id_out):
-        AbstractSubTask.__init__(self, settings_obj, 'DPPP')
+        AbstractSubTask.__init__(self, settings_obj, 'DP3')
         self.db = database.Database(self.settings)
         self.ms_files = ms_files
         self.obs_id_out = obs_id_out
@@ -425,7 +554,8 @@ class MergeMSSubTask(AbstractSubTask):
         return os.path.join(self.get_dir(), f'{self.obs_id_out}.MS')
 
     def get_log_file(self):
-        return os.path.join(self.get_out_file().replace('.MS', '.log'))
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.get_out_file().replace('.MS', f'_{date_str}.log'))
 
     def make_link(self):
         inp_ms_dir = os.path.join(self.get_dir(), 'input_ms')
@@ -453,7 +583,7 @@ class MergeMSSubTask(AbstractSubTask):
         merge_ms = self.settings.merge_ms
 
         steps = ['filter']
-        args = [f'msin={inp_ms_dir}/*.MS', f'msin.datacolumn={merge_ms.data_col}']
+        args = [f'msin={inp_ms_dir}/*.MS', f'msin.datacolumn={merge_ms.data_col}', f'numthreads={merge_ms.numthreads}']
         args.append(f'msout={self.get_out_file()} msout.overwrite=true')
         args.append(f'filter.type=filter filter.blrange=[{merge_ms.blmin},{merge_ms.blmax}]')
 
@@ -465,7 +595,7 @@ class MergeMSSubTask(AbstractSubTask):
 
         if merge_ms.apply_aoflagger:
             steps.append('flagger')
-            args.append('flagger.type=aoflagger flagger.memoryperc=30')
+            args.append(f'flagger.type=aoflagger flagger.memoryperc={merge_ms.memoryperc} flagger.strategy="{merge_ms.aoflagger_strategy}"')
 
         if merge_ms.flag_baselines:
             steps.append('preflag')
@@ -520,6 +650,102 @@ class MergeMSTask(AbstractTask):
         return all_obs_id_out
 
 
+class AOFlaggerSubTask(AbstractSubTask):
+
+    def __init__(self, settings_obj, ms_file):
+        AbstractSubTask.__init__(self, settings_obj, 'DP3')
+        self.ms_file = ms_file
+
+    def get_log_file(self):
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.ms_file.replace('.MS', f'_aoflag_{date_str}.log'))
+
+    def get_parameters(self):
+        ao_flag = self.settings.ao_flag
+        args = [f'msin={self.ms_file}', 'msout=.', f'msin.datacolumn={ao_flag.data_col}', 'steps=[flagger]', f'numthreads={ao_flag.numthreads}']
+        args.append(f'flagger.type=aoflagger flagger.memoryperc={ao_flag.memoryperc} flagger.strategy="{ao_flag.aoflagger_strategy}"')
+
+        return args
+
+
+class AOFlaggerTask(AbstractTask):
+
+    name = 'ao_flag'
+    desc = 'Run AO flagger on the Measurement Set(s)'
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'AOFlagger', settings_obj, max_concurrent=1)
+        self.db = database.Database(self.settings)
+
+    def run(self, obs_ids):
+        all_obs_id_out = []
+
+        for obs_id in obs_ids:
+            if not self.db.has_obs_id(obs_id):
+                print(f'Error: Obs ID {obs_id} does not exist.')
+                return []
+
+            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
+
+            if not ms_files:
+                print(f'Error: Obs ID {obs_id} has no MS files registered.')
+                return []
+
+            for ms_file in np.array(ms_files).flatten():
+                self.add_sub_task(AOFlaggerSubTask(self.settings, ms_file))
+            all_obs_id_out.append(obs_id)
+
+        self.execute()
+
+        return all_obs_id_out
+
+
+class DP3SubTask(AbstractSubTask):
+
+    def __init__(self, settings_obj, ms_file):
+        AbstractSubTask.__init__(self, settings_obj, 'DP3')
+        self.ms_file = ms_file
+
+    def get_log_file(self):
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.ms_file.replace('.MS', f'_dp3_{date_str}.log'))
+
+    def get_parameters(self):
+        return [self.settings.dp3.dp3_file, f'msin={self.ms_file}']
+
+
+class DP3Task(AbstractTask):
+
+    name = 'dp3'
+    desc = 'Run DP3 on the Measurement Set(s)'
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'DP3', settings_obj, max_concurrent=1)
+        self.db = database.Database(self.settings)
+
+    def run(self, obs_ids):
+        all_obs_id_out = []
+
+        for obs_id in obs_ids:
+            if not self.db.has_obs_id(obs_id):
+                print(f'Error: Obs ID {obs_id} does not exist.')
+                return []
+
+            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
+
+            if not ms_files:
+                print(f'Error: Obs ID {obs_id} has no MS files registered.')
+                return []
+
+            for ms_file in np.array(ms_files).flatten():
+                self.add_sub_task(DP3SubTask(self.settings, ms_file))
+            all_obs_id_out.append(obs_id)
+
+        self.execute()
+
+        return all_obs_id_out
+
+
 class MakePSSubTask(AbstractSubTask):
 
     def __init__(self, settings_obj, obs_id, lst_idx=None):
@@ -529,7 +755,8 @@ class MakePSSubTask(AbstractSubTask):
         self.vis_rev = database.VisRevision(self.settings)
 
     def get_log_file(self):
-        return os.path.join(self.vis_rev.get_h5_path(self.obs_id, self.lst_idx), 'logs', 'make_ps.log')
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.vis_rev.get_h5_path(self.obs_id, self.lst_idx), 'logs', f'make_ps_{date_str}.log')
 
     def get_parameters(self):
         h5_i = self.vis_rev.get_h5_file(self.obs_id, 'I', lst_idx=self.lst_idx)
@@ -587,20 +814,28 @@ class RunGPRSubTask(AbstractSubTask):
         self.vis_rev = database.VisRevision(self.settings)
 
     def get_log_file(self):
-        return os.path.join(self.vis_rev.get_path(), self.obs_id, 'logs', 'run_gpr.log')
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.vis_rev.get_path(), self.obs_id, 'logs', f'run_gpr_{date_str}.log')
 
     def get_parameters(self):
         h5_i = self.vis_rev.get_h5_file(self.obs_id, 'I', lst_idx=self.lst_idx)
         h5_v = self.vis_rev.get_h5_file(self.obs_id, 'V', lst_idx=self.lst_idx)
         psc = self.settings.power_spectra
         gpr = self.settings.gpr
-        out_dir = os.path.join(self.vis_rev.get_h5_path(self.obs_id, self.lst_idx))
-        plot_out_dir = os.path.join(self.vis_rev.get_result_path(self.obs_id, self.lst_idx))
+        out_dir = os.path.join(self.vis_rev.get_h5_path(self.obs_id, self.lst_idx), gpr.name)
+        plot_out_dir = os.path.join(self.vis_rev.get_result_path(self.obs_id, self.lst_idx), gpr.name)
 
         args = [h5_i, h5_v, gpr.get_path("config_i"), gpr.get_path("config_v")]
         args.append(f'-f {psc.get_path("flagger")} -e {psc.get_path("eor_bin_list")}')
         args.append(f'-c {psc.get_path("ps_config")} -o {out_dir} -po {plot_out_dir}')
 
+        if gpr.use_v_dt_as_noise:
+            h5_v_dt = self.vis_rev.get_h5_file(self.obs_id, 'dt_V', lst_idx=self.lst_idx)
+            args.append(f'--noise_cube {h5_v_dt}')
+
+        if not gpr.plot_results:
+            args.append('--no_plot')
+
         return args
 
 
@@ -637,6 +872,166 @@ class RunGPRTask(AbstractTask):
         return obs_ids
 
 
+class RunMLGPRSubTask(AbstractSubTask):
+
+    def __init__(self, settings_obj, obs_id, lst_idx=None):
+        AbstractSubTask.__init__(self, settings_obj, 'pstool run_ml_gpr')
+        self.obs_id = obs_id
+        self.lst_idx = lst_idx
+        self.vis_rev = database.VisRevision(self.settings)
+
+    def get_log_file(self):
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.vis_rev.get_path(), self.obs_id, 'logs', f'run_ml_gpr_{date_str}.log')
+
+    def get_parameters(self):
+        psc = self.settings.power_spectra
+        gpr = self.settings.ml_gpr
+
+        h5_i = self.vis_rev.get_h5_file(self.obs_id, gpr.stokes_i, lst_idx=self.lst_idx)
+        h5_v = self.vis_rev.get_h5_file(self.obs_id, gpr.stokes_v, lst_idx=self.lst_idx)
+
+        out_dir = os.path.join(self.vis_rev.get_h5_path(self.obs_id, self.lst_idx), gpr.name)
+        plot_out_dir = os.path.join(self.vis_rev.get_result_path(self.obs_id, self.lst_idx), gpr.name)
+
+        args = [h5_i, h5_v, gpr.get_path("config")]
+        args.append(f'-f {psc.get_path("flagger")} -e {psc.get_path("eor_bin_list")}')
+        args.append(f'-c {psc.get_path("ps_config")} -o {out_dir} -po {plot_out_dir}')
+
+        if gpr.use_v_dt_as_noise:
+            h5_v_dt = self.vis_rev.get_h5_file(self.obs_id, gpr.stokes_dt, lst_idx=self.lst_idx)
+            args.append(f'--file_noise {h5_v_dt}')
+
+        if not gpr.plot_results:
+            args.append('--no_plot')
+
+        return args
+
+
+class RunMLGPRTask(AbstractTask):
+
+    name = 'run_ml_gpr'
+    desc = 'Run ML-GPR and produce power-spectra'
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'ML-GPR', settings_obj)
+        self.vis_rev = database.VisRevision(self.settings)
+        self.img_rev = database.ImageRevision(self.settings)
+
+    def run(self, obs_ids):
+        if not (self.settings.validate('power_spectra') and self.settings.validate('gpr')):
+            return []
+
+        if not self.vis_rev.exist():
+            print(f'Error: visibility revision does not exist.')
+            return []
+
+        for obs_id in obs_ids:
+            if not self.vis_rev.has_obs_id(obs_id):
+                print(f'Error: visibility revision for obs ID {obs_id} not available')
+                return []
+
+            all_lst_idx = self.img_rev.get_lst_bins(obs_id).keys() if self.settings.image.lst_bins else [None]
+
+            for lst_idx in all_lst_idx:
+                self.add_sub_task(RunMLGPRSubTask(self.settings, obs_id, lst_idx))
+
+        self.execute()
+
+        return obs_ids
+
+
+class RunMLGPRInjSubTask(AbstractSubTask):
+
+    def __init__(self, settings_obj, obs_id, var_21_over_noise, x1, x2, lst_idx=None):
+        AbstractSubTask.__init__(self, settings_obj, 'pstool run_ml_gpr_inj')
+        self.obs_id = obs_id
+        self.lst_idx = lst_idx
+        self.var_21_over_noise = var_21_over_noise
+        self.x1 = x1
+        self.x2 = x2
+        self.vis_rev = database.VisRevision(self.settings)
+
+    def get_log_file(self):
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.vis_rev.get_path(), self.obs_id, 'logs', f'run_ml_gpr_inj_{date_str}.log')
+
+    def get_parameters(self):
+        h5_i = self.vis_rev.get_h5_file(self.obs_id, 'I', lst_idx=self.lst_idx)
+        h5_v = self.vis_rev.get_h5_file(self.obs_id, 'V', lst_idx=self.lst_idx)
+        psc = self.settings.power_spectra
+        gpr = self.settings.ml_gpr
+        out_dir = os.path.join(self.vis_rev.get_h5_path(self.obs_id, self.lst_idx), gpr.name)
+
+        args = [h5_i, h5_v, gpr.get_path("config")]
+        args.append(f'-f {psc.get_path("flagger")} -e {psc.get_path("eor_bin_list")}')
+        args.append(f'-c {psc.get_path("ps_config")} -o {out_dir}')
+
+        if gpr.use_v_dt_as_noise:
+            h5_v_dt = self.vis_rev.get_h5_file(self.obs_id, 'dt_V', lst_idx=self.lst_idx)
+            args.append(f'--file_noise {h5_v_dt}')
+
+        args.append(f'-a {self.var_21_over_noise} --x1 {self.x1} --x2 {self.x2}')
+
+        if not gpr.plot_results:
+            args.append('--no_plot')
+
+        return args
+
+
+class RunMLGPRInjTask(AbstractTask):
+
+    name = 'run_ml_gpr_inj'
+    desc = 'Run ML-GPR injection and produce power-spectra'
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'ML-GPR', settings_obj)
+        self.vis_rev = database.VisRevision(self.settings)
+        self.img_rev = database.ImageRevision(self.settings)
+
+    def run(self, obs_ids):
+        if not (self.settings.validate('power_spectra') and self.settings.validate('gpr')):
+            return []
+
+        if not self.vis_rev.exist():
+            print(f'Error: visibility revision does not exist.')
+            return []
+
+        for obs_id in obs_ids:
+            if not self.vis_rev.has_obs_id(obs_id):
+                print(f'Error: visibility revision for obs ID {obs_id} not available')
+                return []
+
+            all_lst_idx = self.img_rev.get_lst_bins(obs_id).keys() if self.settings.image.lst_bins else [None]
+
+            for lst_idx in all_lst_idx:
+                var = self.settings.ml_gpr_inj.var_21_over_noise
+                x1s = self.settings.ml_gpr_inj.x1
+                x2s = self.settings.ml_gpr_inj.x2
+                x1_x2_file = self.settings.ml_gpr_inj.x1_x2_file
+
+                try:
+                    if x1_x2_file:
+                        x1_x2s = np.loadtxt(x1_x2_file, delimiter=',').tolist()
+                    else:
+                        x1_x2s = list(itertools.product(x1s, x2s))
+                except IOError as e:
+                    print(f"Error loading file '{x1_x2_file}': {e}")
+                    return []
+                except ValueError as e:
+                    print(f"Error parsing file '{x1_x2_file}': {e}")
+                    return []
+
+                for v, (x1, x2) in itertools.product(var, x1_x2s):
+                    self.add_sub_task(RunMLGPRInjSubTask(self.settings, obs_id, v, x1, x2, lst_idx))
+
+                self.add_sub_task(RunMLGPRInjSubTask(self.settings, obs_id, 0, 0, 0, lst_idx))
+
+        self.execute()
+
+        return obs_ids
+
+
 class CombineTask(AbstractTask):
 
     name = 'combine'
@@ -661,7 +1056,10 @@ class CombineTask(AbstractTask):
             for obs_id in obs_ids:
                 h5_i = self.vis_rev.get_h5_file(obs_id, 'I')
                 h5_v = self.vis_rev.get_h5_file(obs_id, 'V')
-                h5_dt = self.vis_rev.get_h5_file(obs_id, 'dt_V')
+                if self.settings.combine.use_v_as_dt:
+                    h5_dt = h5_v
+                else:
+                    h5_dt = self.vis_rev.get_h5_file(obs_id, 'dt_V')
 
                 f.write(f"{obs_id} {h5_i} {h5_v} {h5_dt}\n")
 
@@ -720,7 +1118,8 @@ class CombineTask(AbstractTask):
         umax = self.settings.vis_cube.umax
         c = self.settings.combine
 
-        out_log = os.path.join(self.get_path(), f'C{obs_id_out_prefix}.log')
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        out_log = os.path.join(self.get_path(), f'C{obs_id_out_prefix}_{date_str}.log')
 
         if is_sph:
             cmd = (f'pstool combine_sph {self.get_combine_file()} -o {output_template} '
@@ -826,6 +1225,10 @@ class ApplyFlagSubTask(AbstractSubTask):
         self.ms_file = ms_file
         self.h5_file = h5_file
 
+    def get_log_file(self):
+        date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+        return os.path.join(self.ms_file.replace('.MS', f'apply_flags_{date_str}.log'))
+
     def get_parameters(self):
         return [self.ms_file, self.h5_file]
 
@@ -903,26 +1306,31 @@ class RestoreFlagTask(AbstractTask):
 class DelayFlaggerSubTask(AbstractSubTask):
 
     def __init__(self, settings_obj, ms_file):
-        AbstractSubTask.__init__(self, settings_obj, 'flagtool delay_flag')
+        AbstractSubTask.__init__(self, settings_obj, 'flagtool delay_flagger')
         self.ms_file = ms_file
+        self.date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
 
     def get_log_file(self):
-        return os.path.join(self.ms_file.replace('.MS', '_delay_flagger.log'))
+        return os.path.join(self.ms_file.replace('.MS', f'_delay_flagger_{self.date_str}.log'))
 
     def get_path(self):
-        d = os.path.join(os.path.dirname(self.ms_file), 'flagger')
+        d = os.path.join(os.path.dirname(self.ms_file), f'flagger_{self.date_str}')
         futils.mkdir(d)
         return d
 
     def get_parameters(self):
         s = self.settings.delay_flagger
-        return [self.ms_file, f'--data_col={self.settings.image.data_col}', f'--umin={s.umax}', f'--umin={s.umax}',
+        return [self.ms_file, f'--data_col={self.settings.image.data_col}', f'--umin={s.umin}', f'--umax={s.umax}',
                 f'--n_time_avg={s.n_time_avg}', f'--n_times={s.n_times}', f'--n_sigma_i={s.n_sigma_i}',
-                f'--n_sigma_v={s.n_sigma_v}', '--backup', f'--backup_file={H5_FLAG_BACKUP}']
+                f'--n_sigma_v={s.n_sigma_v}', '--backup', f'--backup_file={H5_FLAG_BACKUP}',
+                f'--plot_dir={self.get_path()}']
 
 
 class DelayFlagger(AbstractTask):
 
+    name = 'delay_flagger'
+    desc = 'Apply delay flagger on MS'
+
     def __init__(self, settings_obj):
         AbstractTask.__init__(self, 'DelayFlagger', settings_obj)
         self.db = database.Database(self.settings)
@@ -950,6 +1358,55 @@ class DelayFlagger(AbstractTask):
         return obs_ids
 
 
+class VisFlaggerSubTask(AbstractSubTask):
+
+    def __init__(self, settings_obj, ms_file):
+        AbstractSubTask.__init__(self, settings_obj, 'flagtool vis_flagger')
+        self.ms_file = ms_file
+        self.date_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
+
+    def get_log_file(self):
+        return os.path.join(self.ms_file.replace('.MS', f'{self.date_str}_vis_flagger.log'))
+
+    def get_path(self):
+        d = os.path.join(os.path.dirname(self.ms_file), f'vis_flagger_{self.date_str}')
+        futils.mkdir(d)
+        return d
+
+    def get_parameters(self):
+        return [self.ms_file, self.settings.vis_flagger.config_file, f'--plot_dir={self.get_path()}',
+                '--backup', f'--backup_file={H5_FLAG_BACKUP}']
+
+
+class VisFlagger(AbstractTask):
+
+    name = 'vis_flagger'
+    desc = 'Apply vis flagger on MS'
+
+    def __init__(self, settings_obj):
+        AbstractTask.__init__(self, 'VisFlagger', settings_obj)
+        self.db = database.Database(self.settings)
+
+    def run(self, obs_ids):
+        for obs_id in obs_ids:
+            if not self.db.has_obs_id(obs_id):
+                print(f'Error: Obs ID {obs_id} does not exist.')
+                return []
+
+            ms_files = self.db.get_obs_id_meta(obs_id).ms_files
+
+            if not ms_files:
+                print(f'Error: Obs ID {obs_id} has no MS files registered.')
+                return []
+
+            for ms_file in np.array(ms_files).flatten():
+                self.add_sub_task(VisFlaggerSubTask(self.settings, ms_file))
+
+        self.execute()
+
+        return obs_ids
+
+
 class SSINSSubTask(AbstractSubTask):
 
     def __init__(self, settings_obj, ms_file):
