@@ -9,6 +9,8 @@ import click
 from tabulate import tabulate
 from click import confirm, style
 
+import numpy as np
+
 from libpipe import futils
 from pspipe import __version__
 from pspipe import settings, database
@@ -45,16 +47,21 @@ def init(template_name, data_dir, rev_name, config_modifier):
     ''' Initialize new configuration
 
     \b
-    TEMPLATE_NAME: Template name (nenufar, hba or a12)
+    TEMPLATE_NAME: Template name (nenufar, hba, a12 or alo)
     DATA_DIR: Data directory
     REV_NAME: Image revision name
     CONFIG_MODIFIER: Optional configuration modifiers, of the form key=value
     '''
+    if os.listdir(os.getcwd()):
+        print('Current working directory is not empty, configuration file can not be created.')
+        return
+
     settings_file = settings.get_template_path(template_name)
     print(f'Initializing configuration from {settings_file}')
 
     rev_name = rev_name.replace('.toml', '')
     conf_file = rev_name + '.toml'
+
     s = load_settings(settings_file, config_modifier, **{'data_dir': futils.abspath(data_dir),
                                                          'image': {'name': rev_name}})
     s.duplicate(conf_file, copy_parset_files=True)
@@ -78,7 +85,7 @@ def clone(settings_file, data_dir, config_modifier):
 
     db_in = database.Database(s_in)
     db_out = database.Database(s_out)
-    for obs_id in db_out.list_obs_ids():
+    for obs_id in db_in.list_obs_ids():
         db_out.new_obs_id_from_other(obs_id, db_in.get_obs_id_meta(obs_id))
         print(f'Adding obs_sd {obs_id}')
 
@@ -172,13 +179,23 @@ def list_all_obs(settings_file):
     print_obs_ids(db, db.list_obs_ids())
 
 
+def contains_duplicates(X):
+    return len(np.unique(X)) != len(X)
+
+
 def add(db, obs_id, ms_file_list, **kargs):
     if db.has_obs_id(obs_id):
         print(f'Error: obs_id {obs_id} already exist.')
         return
 
     if ms_file_list:
-        ms_files = [k.strip().split() for k in open(ms_file_list).readlines()]
+        ms_files = [k.strip().split() for k in open(ms_file_list).readlines() if k.strip()]
+        flat_ms_files = [os.path.basename(os.path.normpath(ms_file)) for sublist in ms_files for ms_file in sublist]
+
+        if contains_duplicates(flat_ms_files):
+            print('Error: All MS should have different names')
+            return
+
         db.new_obs_id(obs_id, ms_files)
     else:
         db.new_obs_id(obs_id, **kargs)
@@ -190,8 +207,8 @@ def add(db, obs_id, ms_file_list, **kargs):
 @click.argument('settings_file', type=t_file)
 @click.argument('obs_id', type=str)
 @click.option('--ms_file_list', '-m', type=t_file, help='MS file list')
-@click.option('--int_time', '-i', type=float, help='Integration time', default=0)
-@click.option('--total_time', '-t', type=float, help='Total time', default=0)
+@click.option('--int_time', '-i', type=float, help='Integration time (in second)', default=0)
+@click.option('--total_time', '-t', type=float, help='Total time (in second)', default=0)
 @click.option('--chan_width', '-w', type=float, help='Channel width in Hz', default=0)
 def add_obs(settings_file, obs_id, ms_file_list, int_time, total_time, chan_width):
     ''' Add obs_ids to the given pspipe database
@@ -259,11 +276,11 @@ def set_obs_info(settings_file, obs_ids, ms_file_list, int_time, total_time, cha
 @click.option('--even', '-e', help='Fits images are even set', is_flag=True)
 @click.option('--odd', '-o', help='Fits images are odd set', is_flag=True)
 def add_rev_img(settings_file, obs_id, fits_images, even, odd):
-    ''' Add FITS_IMAGES to the given pspipe database
+    ''' Add FITS_IMAGES to the given pspipe database. An OBS_ID is required, which can be created with the 'add_obs' task.
     '''
     db = database.Database(load_settings(settings_file))
     if not db.has_obs_id(obs_id):
-        print(f'Error: obs_id {obs_id} does not exist. Create with psdb add {obs_id}')
+        print(f'Error: obs_id {obs_id} does not exist. Create with psdb add_obs {obs_id}')
         return
 
     if db.has_image_revisions(db.settings.image.name):
@@ -379,5 +396,59 @@ def remove_obs(settings_file, obs_ids):
         print('No changes made.')
 
 
+def concatenate_datacubes(cubes):
+    from ps_eor import datacube
+
+    key_fct = lambda a: a.freqs[0]
+    cubes = sorted(cubes, key=key_fct)
+
+    assert len(cubes) > 1
+    assert np.alltrue([isinstance(cube, datacube.DataCube) for cube in cubes])
+    assert np.alltrue([np.allclose(cubes[0].uu, cube.uu) for cube in cubes[1:]])
+    assert np.alltrue([np.allclose(cubes[0].vv, cube.vv) for cube in cubes[1:]])
+
+    data = np.concatenate([c.data for c in cubes])
+    weights_data = np.concatenate([c.weights.data for c in cubes])
+    freqs = np.concatenate([c.freqs for c in cubes])
+    weights = datacube.CartWeightCube(weights_data, cubes[0].uu, cubes[0].vv, freqs, cubes[0].meta)
+
+    return datacube.CartDataCube(data, cubes[0].uu, cubes[0].vv, freqs, cubes[0].meta, weights=weights)
+
+
+@main.command('concat_vis_cube')
+@click.argument('settings_file', type=t_file)
+@click.argument('obs_ids', type=str)
+@click.argument('new_obs_id', type=str)
+@click.option('--fmin', type=float, help='Min frequency in MHz', default=None)
+@click.option('--fmax', type=float, help='Max frequency in MHz', default=None)
+def concat_vis_cube(settings_file, obs_ids, new_obs_id, fmin, fmax):
+    ''' Remove an Obs ID from the given pspipe database
+    '''
+    rev = database.VisRevision(load_settings(settings_file))
+    db = database.Database(load_settings(settings_file))
+    obs_id_set = list(db.filter_obs_ids(obs_ids))
+
+    if not obs_id_set:
+        print(f'Error: No obs_ids matching {obs_ids} in database.')
+        return
+
+    meta = db.get_obs_id_meta(obs_id_set[0])
+
+    add(db, new_obs_id, None, int_time=meta.int_time, total_time=meta.total_time, chan_width=meta.chan_width)
+    data1 = rev.get_data(obs_id_set[0])
+
+    all_stokes = [s for s in data1.all_stokes.values() if data1.has_stokes(s)]
+
+    for s in all_stokes:
+        data_cube = concatenate_datacubes([rev.get_data(obs_id).get_stokes(s) for obs_id in obs_id_set])
+        if fmin is not None and fmax is not None:
+            data_cube = data_cube.get_slice(fmin, fmax)
+        h5_filename = rev.get_h5_file(new_obs_id, s)
+        os.makedirs(os.path.dirname(h5_filename),  exist_ok=True)
+        data_cube.save(h5_filename)
+
+    print(f'Combined data for {new_obs_id} successfully added.')
+
+
 if __name__ == '__main__':
     main()
